import numpy as np

from .batch_mean import batch_mean

def bm_cov(x, b):
    col_batch_means = np.apply_along_axis(lambda v: batch_mean(v, b), 0, x)
    return b * np.cov(col_batch_means, rowvar=False)



import numpy as np

from .mc_cov import mc_cov

def multi_ess(x, mc_cov_mat=None, method='inse', adjust=False, b=None, r=3):
    num_iters, num_pars = x.shape

    cov_mat_det = np.linalg.det(np.cov(x, rowvar=False))
    mc_cov_mat_det = np.linalg.det(
        mc_cov(x, method=method, adjust=adjust, b=b, r=r, rowvar=False) if mc_cov_mat is None else mc_cov_mat
    )

    return num_iters * ((cov_mat_det / mc_cov_mat_det) ** (1/num_pars))



from .acf import acf
from .cmaps import redblue_cmap
from .cor_heatmap import cor_heatmap
from .hist import hist
from .hist2d import hist2d
from .running_mean import running_mean
from .trace import trace



import numpy as np

from matplotlib import cm
from matplotlib.colors import ListedColormap

redblue_cmap = ListedColormap(
    np.vstack((cm.get_cmap('Blues', 128)(np.linspace(1, 0, 128)), cm.get_cmap('Reds', 128)(np.linspace(0, 1, 128)))),
    name='RedBlue'
)



import numpy as np

from pathlib import Path

import kanga.stats as st

class ChainArray:
    def __init__(self, vals):
        self.reset(vals)

    def reset(self, vals):
        self.vals = vals

    @classmethod
    def from_file(selfclass, keys=['sample', 'target_val', 'accepted'], path=Path.cwd(), dtype=np.float64):
        vals = {}

        for key in keys:
            vals[key] = np.loadtxt(Path(path).joinpath(key+'.csv'), dtype=int if key == 'accepted' else dtype, delimiter=',')

        return selfclass(vals)

    def __repr__(self):
        return f"Markov chain containing {len(self.vals['sample'])} samples."

    def __len__(self):
        return self.num_samples()

    def num_params(self):
        return len(self.get_sample(0))

    def num_samples(self):
        return len(self.get_samples())

    def get_param(self, idx):
        return self.vals['sample'][:, idx]

    def get_sample(self, idx):
        return self.vals['sample'][idx, :]

    def get_samples(self):
        return self.vals['sample']

    def get_target_vals(self):
        return self.vals['target_val']

    def get_grad_val(self, idx):
        return self.vals['grad_val'][idx, :]

    def get_grad_vals(self):
        return self.vals['grad_val']

    def state(self, idx=-1):
        current = {}
        for key, val in self.vals.items():
            try:
                current[key] = val[idx]
            except IndexError:
                print(f'WARNING: chain does not have values for {key}.')
                pass
        return current

    def mean(self):
        return self.get_samples().mean(0)

    def running_mean(self, idx):
        return np.array(st.running_mean(self.get_param(idx)))

    def running_means(self):
        return np.transpose(np.array([self.running_mean(i) for i in range(self.num_params())]))

    def mc_se(self, mc_cov_mat=None, method='inse', adjust=False, b=None, r=3):
        if mc_cov_mat is None:
            return st.mc_se(self.get_samples(), method=method, adjust=adjust, b=b, r=r, rowvar=False)
        else:
            return st.mc_se_from_cov(mc_cov_mat)

    def mc_cov(self, method='inse', adjust=False, b=None, r=3):
        return st.mc_cov(self.get_samples(), method=method, adjust=adjust, b=b, r=r, rowvar=False)

    def mc_cor(self, mc_cov_mat=None, method='inse', adjust=False, b=None, r=3):
        if mc_cov_mat is None:
            return st.mc_cor(self.get_samples(), method=method, adjust=adjust, b=b, r=r, rowvar=False)
        else:
            return st.cor_from_cov(mc_cov_mat)

    def acceptance_rate(self):
        return sum(self.vals['accepted']) / self.num_samples()

    def block_acceptance_rate(self):
        return np.array(self.vals['accepted']).sum(axis=0) / self.num_samples()

    def multi_ess(self, mc_cov_mat=None, method='inse', adjust=False, b=None, r=3):
        return st.multi_ess(self.get_samples(), mc_cov_mat=mc_cov_mat, method=method, adjust=adjust, b=b, r=r)

    def savecsv(self, keys=None, path=Path.cwd(), fmt=None):
        keys = keys or self.vals.keys()
        fmt = fmt or ['%.18e' for _ in range(len(keys))]

        for i, key in enumerate(keys):
            np.savetxt(path.joinpath(key+'.csv'), self.vals[key], fmt=fmt[i], delimiter=',')



import numpy as np

from kanga.emulators import approx_ar_cov_sum

def optimal_ar_batch_size(x):
    num_iters = x.shape[0]

    ar_approximation = np.apply_along_axis(approx_ar_cov_sum, 0, x) ** 2
    coeff = (ar_approximation[0, ].sum() / ar_approximation[1, ].sum()) ** (1 / 3)

    b = int(np.floor((num_iters ** (1 / 3)) * coeff))
    if b == 0:
        b = 1

    return b



from statsmodels.tsa.stattools import acovf

from .fit_ar import fit_ar

def approx_ar_cov_sum(x):
    params, sigma2, order = fit_ar(x)
    param_sum = params.sum()
    asympt_var = sigma2 / ((1 - param_sum) ** 2)

    gammas = acovf(x, nlag=order-1, fft=False)

    if (order != 0):
        Gamma = 0

        for i in range(1, order+1):
            for k in range(1, i+1):
                Gamma = Gamma + params[i-1] * k * gammas[i-k]

        # Gamma is computed using the equation at the bottom of p. 9 in https://arxiv.org/pdf/1804.05975.pdf
        # See also https://stats.stackexchange.com/questions/371792/sum-of-autocovariances-for-arp-model/372006#372006
        Gamma = 2 * (
            Gamma + 0.5 * (asympt_var - gammas[0]) * (params * range(1, order + 1)).sum()
            ) / (1 - param_sum)
    else:
        Gamma = 0

    return Gamma, asympt_var



from .mc_cov import mc_cov
from .mc_se_from_cov import mc_se_from_cov

def mc_se(x, method='inse', adjust=False, b=None, r=3, rowvar=False):
    return mc_se_from_cov(mc_cov(x, method=method, adjust=adjust, b=b, r=r, rowvar=rowvar))



# Implementation taken from
# https://stackoverflow.com/questions/43238173/python-convert-matrix-to-positive-semi-definite/43244194#43244194

import numpy as np

from .is_pos_def import is_pos_def

def nearest_pd(A):
    """Find the nearest positive-definite matrix to input

    A Python/Numpy port of John D'Errico's `nearestSPD` MATLAB code [1], which credits [2]

    [1] https://www.mathworks.com/matlabcentral/fileexchange/42885-nearestspd

    [2] https://doi.org/10.1016/0024-3795(88)90223-6
    """

    B = (A + A.T) / 2
    _, s, V = np.linalg.svd(B)

    H = np.dot(V.T, np.dot(np.diag(s), V))

    A2 = (B + H) / 2

    A3 = (A2 + A2.T) / 2

    if is_pos_def(A3):
        return A3

    spacing = np.spacing(np.linalg.norm(A))
    I = np.eye(A.shape[0])
    k = 1
    while not is_pos_def(A3):
        mineig = np.min(np.real(np.linalg.eigvals(A3)))
        A3 += I * (-mineig * k**2 + spacing)
        k += 1

    return A3



from kanga import stats

from .trace import trace

def running_mean(y, x=None, figsize=[8, 4], linewidth=None, color=None, ylim=None, margins=0., title=None, xlabel=None,
    ylabel=None, axes_labelsize=14, axes_titlesize=14, xticks=None, xticklabels=None, xtick_labelsize=12, yticks=None,
    yticklabels=None, ytick_labelsize=12, fname=None, legend=False, legend_labels=None, legend_loc='best', legend_ncol=None,
    legend_fontsize=12, quality=100, transparent=True, bbox_inches='tight', pad_inches=0.1):
    means = stats.running_mean(y)

    trace(means, x=x, figsize=figsize, linewidth=linewidth, color=color, ylim=ylim, margins=margins, title=title,
        xlabel=xlabel, ylabel=ylabel, axes_labelsize=axes_labelsize, axes_titlesize=axes_titlesize, xticks=xticks,
        xticklabels=xticklabels, xtick_labelsize=xtick_labelsize, yticks=yticks, yticklabels=yticklabels,
        ytick_labelsize=ytick_labelsize, fname=fname, legend=legend, legend_labels=legend_labels, legend_loc=legend_loc,
        legend_ncol=legend_ncol, legend_fontsize=legend_fontsize, quality=quality, transparent=transparent,
        bbox_inches=bbox_inches, pad_inches=pad_inches
    )

    return means
